/* $Workfile: T8pBurst.c $       $Revision: 12 $      $Modtime: 12/14/03 6:07p $  */
/*******************************************************************************

  Odin TeleSystems Inc.

  OTX Telecom Adapter Driver

  @doc PublicDemo

  @module

    T8pBurst.C

    A simple example console application demonstrating the use of the 
    Raw Data Sender and the Raw Data Receiver with the OTX Driver using the 
    board's Burst device to access the data using DMA.

      Thor-8-PCI-Plus
      Sleipnir

    The Demo application provides a simple console interface which utilizes
    simple keystrokes as commands.

    It configures 8 Lines which can be looped with an external loop cable.

    The Raw Data Receiver and the Raw Data Sender is hosted by the
    Burst device.

    Press 'h' or '?' for help. 'q' to quit.

*******************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <assert.h>

#include "OsaStd.h"
#include "OsaCon.h"
#include "OsaThrd.h"
#include "OtxErr.h"
#include "OtxDef.h"
#include "OtxType.h"
#include "OtxDrv.h"
#include "OtxHdlc.h"
#include "OtxRaw.h"
#include "OtxT1E1.h"
#include "OtxData.h"
#include "OtxTone.h"
#include "OtxDsp.h"
#include "OtxBrd.h"

#include "OtxBurst.h"
#include "OtxBurstD.h"

#define DEMO_DEFAULT_LOG_FILE   "T8pBurst.log"
#define DEMO_DEFAULT_BOARD_NO   0

#define DEMO_RAW_TX_BUF_SIZE    2000
#define DEMO_BURST_TX_BUF_SIZE  0x8000      // 32k

#define DEMO_TEST_STRING        "abcdefg"


typedef struct DemoEnvT {
  char            *pszInputStr;
  OTX_BOOL        bExitOnError;
  char            cwd[OTX_MAX_PATH];
  char            logFileName[OTX_MAX_PATH];
  OtxPhysicalDeviceTypeE nBoardType;
  int             coLi;
  OtxT1E1LiModeE  eT1E1Mode;    // T1 or E1 mode
  int             boardNo;
  OTX_INT32       nRawDataCapacity;
  OTX_INT32       nStartingTimeslot;
} DemoEnvT;


/*******************************************************************************
 Function Prototypes
*******************************************************************************/
static void DemoCreateEventQueue(void);
static void DemoCreatePhysicalDevices(void);
static void DemoConfigureE1T1Devices(void);
static void DemoEnablePhysicalDevices(void);
static void DemoCreateLogicalDevices(void);
static void DemoConnectDevices(void);
static void DemoInitializeLogicalDevices(void);
static void DemoEnableLogicalDevices(void);
static void DemoInitT1E1ConfigData(OUT OtxE1ConfigOptionsS *pE1CfgData, OUT OtxT1ConfigOptionsS *pT1CfgData);
static void DemoPrintEvent(IN OtxEventDataS *pEventData);
static void DemoCheckRc(IN OTX_HANDLE hDevice, IN OTX_RESULT otxRc);
static void DemoExitQuestion(void);
static void DemoLogf(char *formatStr, ...);
static void DemoHelpPanel(void);
static void DemoParseCmdLineArgs(int argc, char *argv[], DemoEnvT *envP);


/*******************************************************************************
 Global Data
*******************************************************************************/
OTX_EVENT   g_hNotificationEvent = OTX_INVALID_EVENT_VALUE;
OTX_HANDLE  g_hEventQueue  = 0;
FILE       *g_pDemoLogFile = NULL;
DemoEnvT    g_demoEnv;
OTX_BOOL    g_bHdlcVerbose = OTX_TRUE;
OTX_BOOL    g_bStatsThreadActive = OTX_FALSE;


struct {
  OTX_HANDLE hThorController;
  OTX_HANDLE hLi[8];
  OTX_HANDLE hQuadLi[2];
  OTX_HANDLE hBurst;
} g_hPhysDevs;

struct {
  // Pipes for Raw Data Sender/Receiver in the DSP
  OTX_HANDLE hBurstLiRawData;
  OTX_HANDLE hLiBurstRawData;
} g_hPipes[8];


struct sRawDataSender {
  OTX_HANDLE  hHandle;
  OTX_HANDLE  hEventQueue;
  OTX_EVENT   hNotificationEvent;
  OTX_UINT32  nSourceId;
  OTX_BOOL    bActive;
  OTX_UINT8   sqTxBuf[DEMO_RAW_TX_BUF_SIZE];
  FILE        *pTxFile;
  OTX_INT32   cbTxBuf;
  OTX_INT32   coUnderRunErrors;
};

struct sRawDataReceiver {
  OTX_HANDLE  hHandle;
  OTX_HANDLE  hEventQueue;
  OTX_EVENT   hNotificationEvent;
  OTX_UINT32  nSourceId;
  OTX_BOOL    bVerbose;
  OTX_BOOL    bActive;
  FILE        *pRxFile;
};

struct {
  struct sRawDataSender mRawDataSender[8];
  struct sRawDataReceiver mRawDataReceiver[8];
} g_hLogicDevs;

struct {
  OtxE1ConfigOptionsS e1CfgData;
  OtxT1ConfigOptionsS t1CfgData;
} g_T1E1CfgData;




/*******************************************************************************
*******************************************************************************/


void DemoCreateEventQueue(void)
/*******************************************************************************

  @func

  Create the event queue to be used by the driver to inform the application
  about events, such as line status changes, or detected DMTF digits.

*******************************************************************************/
{
  OTX_RESULT result;

  // Create a notification Event
  result = OtxDrvCreateEvent(&g_hNotificationEvent);
  if (OTX_RESULT_CODE(result) == OTX_E_NOTIMPL) {
    g_hNotificationEvent = NULL;
    result = OTX_S_OK;
  }
  
  DemoCheckRc(OTX_INVALID_HANDLE_VALUE, result);

  // Create Event Queue for driver notifications
  result = OtxDrvCreateEventQueue(128, g_hNotificationEvent, &g_hEventQueue); DemoCheckRc(g_hEventQueue, result);

  // Enable Event Queue
  result = OtxDrvEnable(g_hEventQueue, OTX_TASK_SYNC, 5000L); DemoCheckRc(g_hEventQueue, result);

}



void DemoCreatePhysicalDevices(void)
/*******************************************************************************

  @func

  Create and Connect the Physical Devices on the Thor-8 board

  @comm

    Calls function <f OtxDrvCreatePhysicalDevice> returning a handle to a
    Device Object representing a physical device, which can be the
    entire board or an integrated circuit (chip) on the board. The number of
    physical devices is set per board and cannot change dynamically.

    The physical devices can be classified in 2 categories:

  - BOARD Devices: A board device represents one OTX Network Interface card or
    a Daughter Board Module. The board device is the place holder and controller
    of other physical and sometime logical devices representing
    the devices on the board: such as Line Interface and Filter  Devices.

    The Burst device hosts the DMA functionality.  If we open the Burst
    device, when we enable that device, the process of transferring data from
    the board to the driver will take place via DMA.  The Burst device provides
    the necessary buffering to process the data from the board until it can be
    processed and filtered to the logical devices that are hosted by Burst.
    (see DemoCreateLogicalDevices for example of connection).

  - CHIP Devices: A chip device represents a physical integrated circuit
    on the board performing one or several functions. A logical device can
    be created to represent each of the functions implemented by a chip device.
    For example, a Digital Signal Processor is a physical "chip" device.
    However, the DSP software may implement several functions and be
    represented by a multiple of logical devices.



*******************************************************************************/
{
  OTX_RESULT result;
  OTX_INT32 i;
  OTX_INT32 j;

  // Connect to the Physical Board Controller
  switch (g_demoEnv.nBoardType) {
    case OTX_DEVICE_SLEIPNIR_1_PCI:
      DemoLogf("Opening Sleipnir board\n");
      break;
    case OTX_DEVICE_THOR_8_PCI_PLUS:
      DemoLogf("Opening Thor-8-PCI Plus board\n");
      break;
    default:
      DemoLogf("Board Type (%d) not supported by this demo program\n", g_demoEnv.nBoardType);
      OsaConsoleExit();
      exit(1);
      break;
  }

  result = OtxDrvOpenPhysicalDevice(0, g_demoEnv.nBoardType, g_demoEnv.boardNo, g_demoEnv.boardNo, g_hEventQueue, &g_hPhysDevs.hThorController); DemoCheckRc(g_hPhysDevs.hThorController, result);

  if (g_demoEnv.nBoardType == OTX_DEVICE_THOR_8_PCI_PLUS) {
    for (i = 0; i < 2; i++) {
      // Get handles to the two quad T1/E1 line interface devices
      result = OtxDrvOpenPhysicalDevice(g_hPhysDevs.hThorController, OTX_DEVICE_QUAD_T1E1, i, i, g_hEventQueue, &g_hPhysDevs.hQuadLi[i]); DemoCheckRc(g_hPhysDevs.hThorController, result);

      // Get handles to the 4 individual T1/E1 line interface members within the Quad device
      for (j = 0; j < 4; j++) {
        result = OtxDrvOpenPhysicalDevice(g_hPhysDevs.hQuadLi[i], OTX_DEVICE_LI_T1E1, j, j + i*4, g_hEventQueue, &g_hPhysDevs.hLi[j + i*4]); DemoCheckRc(g_hPhysDevs.hQuadLi[i], result);
      }
    }

  } else if (g_demoEnv.nBoardType == OTX_DEVICE_SLEIPNIR_1_PCI) {

    // Open the Line interfaces
    for (i = 0; i < g_demoEnv.coLi; i++) {
      result = OtxDrvOpenPhysicalDevice(g_hPhysDevs.hThorController, OTX_DEVICE_LI_T1E1, i, i, g_hEventQueue, &g_hPhysDevs.hLi[i]); DemoCheckRc(g_hPhysDevs.hThorController, result);
    }

    // we'll source the Falc from the Burst device, and source the burst device from the FALC
    //  Falc <-- Burst Send (P2S)
    result = OtxDrvSetAttributeValueUint32(g_hPhysDevs.hThorController, OTX_ATTR_BRD_FALC_SOURCED, OTX_BRD_FALC_SOURCED_P2S); DemoCheckRc(g_hPhysDevs.hThorController, result);
    //  Burst Recv (S2P) <-- Falc
    result = OtxDrvSetAttributeValueUint32(g_hPhysDevs.hThorController, OTX_ATTR_BRD_S2P_SOURCED, OTX_BRD_S2P_SOURCED_FALC); DemoCheckRc(g_hPhysDevs.hThorController, result);
  }

  // Get a handle to the physical Burst device.
  result = OtxDrvOpenPhysicalDevice(g_hPhysDevs.hThorController, OTX_DEVICE_BURST, 0, 0, g_hEventQueue, &g_hPhysDevs.hBurst); DemoCheckRc(g_hPhysDevs.hThorController, result);
}



/*******************************************************************************

  @func
    Configure E1/T1 Devices

  @comm The clock source of the board(s) must set by a call 
  to <f OtxBrdSetClocks> prior to calling this function.

*******************************************************************************/
static void DemoConfigureE1T1Devices(void)
{
	OTX_RESULT result;
  OTX_INT32  i;
  
  // Set default configuration data
  DemoInitT1E1ConfigData(&(g_T1E1CfgData.e1CfgData), &(g_T1E1CfgData.t1CfgData));

	if (g_demoEnv.eT1E1Mode == OTX_T1E1_MODE_T1){ // T1
		// Configure T1 Transceivers
		for (i = 0; i < g_demoEnv.coLi; i++) {
			result = OtxT1SetConfiguration(g_hPhysDevs.hLi[i], &(g_T1E1CfgData.t1CfgData));  DemoCheckRc(g_hPhysDevs.hLi[i], result);
		}
		// Set them in T1 mode
		for (i = 0; i < g_demoEnv.coLi; i++) {
			result =  OtxT1E1SetLiMode(g_hPhysDevs.hLi[i], OTX_T1E1_MODE_T1); DemoCheckRc(g_hPhysDevs.hLi[i], result);
		}
	}
	else{  // E1
		// Configure E1 Transceivers
		for (i = 0; i < g_demoEnv.coLi; i++) {
			result = OtxE1SetConfiguration(g_hPhysDevs.hLi[i], &(g_T1E1CfgData.e1CfgData));  DemoCheckRc(g_hPhysDevs.hLi[i], result);
		}
		// Set them in E1 mode
		for (i = 0; i < g_demoEnv.coLi; i++) {
			result =  OtxT1E1SetLiMode(g_hPhysDevs.hLi[i], OTX_T1E1_MODE_E1); DemoCheckRc(g_hPhysDevs.hLi[i], result);
		}
	}

}


void DemoEnablePhysicalDevices(void)
/*******************************************************************************

  @func
    Enable Configured Physical Devices

*******************************************************************************/
{
  OTX_RESULT result;
  OTX_INT32 i;

  // Enable the board (Physical Device)
  DemoLogf("Enable g_hPhysDevs.hThorController\n");
  result = OtxDrvEnable(g_hPhysDevs.hThorController,  OTX_TASK_SYNC, 5000L); DemoCheckRc(g_hPhysDevs.hThorController, result);

  if (g_demoEnv.nBoardType == OTX_DEVICE_THOR_8_PCI_PLUS) {
    for (i = 0; i < 2; i++) {
      // Enable QuadFalcs
      DemoLogf("Enable g_hPhysDevs.hQuadLi[%d]\n", i);
      result = OtxDrvEnable(g_hPhysDevs.hQuadLi[i],  OTX_TASK_SYNC, 5000L); DemoCheckRc(g_hPhysDevs.hQuadLi[i], result);
    }

  }

  for (i = 0; i < g_demoEnv.coLi; i++) {
    // Enable Physical Devices, Li's
    DemoLogf("Enable g_hPhysDevs.hLi[%d]\n", i);
    result = OtxDrvEnable(g_hPhysDevs.hLi[i],  OTX_TASK_SYNC, 5000L); DemoCheckRc(g_hPhysDevs.hLi[i], result);
  }

  // by enabling the Burst device, we start the process of transferring data
  // from the hardware to the driver via DMA.  The data is stored in the Burst
  // device's buffer for later processing to connected logical devices.
  DemoLogf("Enable g_hPhysDevs.hBurst\n");
  result = OtxDrvEnable(g_hPhysDevs.hBurst,  OTX_TASK_SYNC, 5000L); DemoCheckRc(g_hPhysDevs.hBurst, result);
}


void DemoCreateLogicalDevices(
/*******************************************************************************

  @func
    Create the needed Logical Devices

  @comm
    The logical devices are differenciated from the phyiscal devices by the fact
    the number of logical devices is not set. More of them can be requested
    by the user and allocated by the system if needed.

    Logical Devices can be classified into Control Device and Data Devices:

    - Control Device:  A logical device performing a control function. E.g. Event Queue.

    - Data Filter Device: A device implemented in Hardware, Firmware of Software
    which converts, processes, or manipulates serial data. Examples of data devices include
    Time-Space Switch, HDLC Controller, Tone Detector, Tone Generator, etc.

    Data Filter devices can be categorized into 3 basic types:

     - Data Source: Originates a Pipe or adds data to an existing Pipe. E.g. DTMF Tone Generator.

     - Data Sink: Terminates a Pipe or extracts data from a Pipe. E.g. DTMF Detector.

     - Data Bridge: Cross-connects data from one Pipe to another or combines data from
     multiple Pipes into one Pipe. E.g. cross-connect in TSS.

*******************************************************************************/
  void
)
{
  OTX_RESULT result;
  OTX_INT32 i;
  //OTX_INT32 j;
  OTX_UINT32 nIrpQueueLimit;
  OTX_UINT32 nFifoSize;

  for (i = 0; i < g_demoEnv.coLi; i++) {

    // Create a notification Event and Event Queue for the Raw Data Sender
    result = OtxDrvCreateEvent(&g_hLogicDevs.mRawDataSender[i].hNotificationEvent); DemoCheckRc(OTX_INVALID_HANDLE_VALUE, result);
    result = OtxDrvCreateEventQueue(128, g_hLogicDevs.mRawDataSender[i].hNotificationEvent, &g_hLogicDevs.mRawDataSender[i].hEventQueue); DemoCheckRc(g_hLogicDevs.mRawDataSender[i].hEventQueue, result);
    result = OtxDrvEnable(g_hLogicDevs.mRawDataSender[i].hEventQueue, OTX_TASK_SYNC, 5000L); DemoCheckRc(g_hLogicDevs.mRawDataSender[i].hEventQueue, result);

    // Create Raw Data Sender
    // The Raw Data Sender is hosted by the Burst device.  Data that is sent
    // via this device is processed into an awaiting buffer in the Burst device
    // and will be DMA'ed to the hardware when the hardware notifies the driver
    // that space is available to transfer outgoing data.
    result = OtxDrvCreateLogicalDevice(g_hPhysDevs.hBurst, OTX_LDEVICE_DATA_RAW_SENDER, g_hLogicDevs.mRawDataSender[i].hEventQueue, i, &g_hLogicDevs.mRawDataSender[i].hHandle); DemoCheckRc(g_hPhysDevs.hBurst, result);
    g_hLogicDevs.mRawDataSender[i].nSourceId = i; // Store sourceId so that we can easily print it from the Sender Thread

    // Increase the IRP Queue limit for fast pipes (to be able to buffer more transmit data in the driver)
    result = OtxDrvGetAttributeValueUint32(g_hLogicDevs.mRawDataSender[i].hHandle, OTX_ATTR_DEVICE_IRP_QUEUE_LIMIT, &nIrpQueueLimit); DemoCheckRc(g_hLogicDevs.mRawDataSender[i].hHandle, result);
    DemoLogf("RawDataSender[%d]: nIrpQueueLimit = %d\n", i, nIrpQueueLimit);
    if (g_demoEnv.nRawDataCapacity > 256) {
      nIrpQueueLimit = 30;
      DemoLogf("RawDataSender[%d]: Setting nIrpQueueLimit = %d\n", i, nIrpQueueLimit);
      result = OtxDrvSetAttributeValueUint32(g_hLogicDevs.mRawDataSender[i].hHandle, OTX_ATTR_DEVICE_IRP_QUEUE_LIMIT, nIrpQueueLimit); DemoCheckRc(g_hLogicDevs.mRawDataSender[i].hHandle, result);
    }

    // Create a notification Event and Event Queue for the Raw Data Receiver
    result = OtxDrvCreateEvent(&g_hLogicDevs.mRawDataReceiver[i].hNotificationEvent); DemoCheckRc(OTX_INVALID_HANDLE_VALUE, result);
    result = OtxDrvCreateEventQueue(128, g_hLogicDevs.mRawDataReceiver[i].hNotificationEvent, &g_hLogicDevs.mRawDataReceiver[i].hEventQueue); DemoCheckRc(g_hLogicDevs.mRawDataReceiver[i].hEventQueue, result);
    result = OtxDrvEnable(g_hLogicDevs.mRawDataReceiver[i].hEventQueue, OTX_TASK_SYNC, 5000L); DemoCheckRc(g_hLogicDevs.mRawDataReceiver[i].hEventQueue, result);

    // Create Raw Data Receiver
    // The Raw Data Receiver is hosted by the Burst device.  As data is received,
    // the data is DMA'ed from the hardware into awaiting buffers inside the
    // burst device.  For each Raw Data Receiver hosted, the Burst device processes
    // its data and places the data destined for the Raw Data Reciever into its buffer,
    // at which the Raw Data Receiver awaits to move the data to this aplication.
    result = OtxDrvCreateLogicalDevice(g_hPhysDevs.hBurst, OTX_LDEVICE_DATA_RAW_RECEIVER, g_hLogicDevs.mRawDataReceiver[i].hEventQueue, i, &g_hLogicDevs.mRawDataReceiver[i].hHandle); DemoCheckRc(g_hPhysDevs.hBurst, result);
    g_hLogicDevs.mRawDataReceiver[i].nSourceId = i; // Store sourceId so that we can easily print it from the Receiver Thread

    // Modify the FIFO size of the Raw Data Receiver
    result = OtxDrvGetAttributeValueUint32(
      g_hLogicDevs.mRawDataReceiver[i].hHandle, 
      OTX_ATTR_RAWDATA_RECEIVER_FIFO_SIZE, 
      &nFifoSize); 
    DemoCheckRc(g_hLogicDevs.mRawDataReceiver[i].hHandle, result);
    DemoLogf("RawDataReceiver[%d]: nFifoSize = %d bytes  (sizePerTimeslot=%d)\n", i, nFifoSize * (g_demoEnv.nRawDataCapacity/64), nFifoSize);
    if (g_demoEnv.nRawDataCapacity > 256) {
      nFifoSize = 2048;
      DemoLogf("RawDataReceiver[%d]: Setting nFifoSize = %d bytes  (sizePerTimeslot=%d)\n", i, nFifoSize * (g_demoEnv.nRawDataCapacity/64), nFifoSize);
      result = OtxDrvSetAttributeValueUint32(
        g_hLogicDevs.mRawDataReceiver[i].hHandle, 
        OTX_ATTR_RAWDATA_RECEIVER_FIFO_SIZE, 
        nFifoSize); 
      DemoCheckRc(g_hLogicDevs.mRawDataReceiver[i].hHandle, result);
    }

    // Prevent the driver from generating the OTX_RAWDATA_RECEIVER_EC_DATA_AVAILABLE event 
    // (to reduce driver overhead)
    result = OtxDrvSetAttributeValueBool(
      g_hLogicDevs.mRawDataReceiver[i].hHandle, 
      OTX_ATTR_RAWDATA_RECEIVER_GENERATE_DATA_EVENTS, 
      OTX_FALSE); 
    DemoCheckRc(g_hLogicDevs.mRawDataReceiver[i].hHandle, result);

  }

}


void DemoConnectDevices(
/*******************************************************************************

  @func
    Create the needed Pipe Devices and connect the physical and logical devices
    to pipes.

  @comm
    A Pipe Device is a logical entity representing a serial bit-stream (a subset of a
    Highway) with a data rate between 8 kbit/s and 2.048Mbit/s. A pipe is a characterized
    bu its capacity (in 8 kbit/s increments). The user can create a pipe with
    the driver function <f OtxDrvCreatePipe>. A pipe is connected to physical or
    logical devices and mapped to specific time-slots with the driver function
    <f OtxDrvConnectPipe>.

*******************************************************************************/
  void
)
{
  OTX_INT32 nLi;
  OTX_RESULT result;
  OTX_UINT32 iHighwayNo;
  OTX_UINT32 cMaxHighwayTS;
  OTX_UINT8 sqTimeSlotMask[32];
  OTX_UINT8 sqBurstTimeSlotMask[OTX_LOCAL_PIPE_TIMESLOTS_MAX];


  // Connect to the Physical Board Controller
  switch (g_demoEnv.nBoardType) {
    case OTX_DEVICE_SLEIPNIR_1_PCI:
      cMaxHighwayTS = 32;
      break;
    case OTX_DEVICE_THOR_8_PCI_PLUS:
      cMaxHighwayTS = 128;
      break;
    default:
      DemoLogf("Board Type (%d) not supported by this demo program\n", g_demoEnv.nBoardType);
      OsaConsoleExit();
      exit(1);
      break;
  }

  // Create the Pipes (one for each direction) between the Burst Device and the LIs
  for (nLi = 0; nLi < g_demoEnv.coLi; nLi++) {
    OTX_INT32 nCapacity;
    OTX_INT32 nTs;

    // we must define the highway to use [0-1]
    iHighwayNo = nLi / 4;

    // Create Raw Data pipes (one for each direction) Use all timeslots except TS0.
    memset(sqTimeSlotMask, 0, sizeof(sqTimeSlotMask) * sizeof(sqTimeSlotMask[0]));
    memset(sqBurstTimeSlotMask, 0, sizeof(sqBurstTimeSlotMask) * sizeof(sqBurstTimeSlotMask[0]));
  
    if (g_demoEnv.nRawDataCapacity >= 64) {

      // Standard (64kpbs) or superchannel
      nCapacity = 0;

      DemoLogf("Creating pipe for RawData Sender/Receiver starting at TS%d\n", g_demoEnv.nStartingTimeslot);

      for (nTs = g_demoEnv.nStartingTimeslot; nTs < 32; nTs++) {
      
        if (nCapacity == g_demoEnv.nRawDataCapacity) {
          break;
        }
      
        if (nTs == 0) {
          continue; // TS0 is sync timeslot
        }
      
        sqTimeSlotMask[nTs] = 0xFF; 
      
        // for this demo, we'll lay out the burst time slot mask as follows:
        //    [ 0..31]  - LI0:TS0-TS31 and LI4:TS0-TS31
        // OTX_DEVICE_THOR_8_PCI_PLUS only:
        //    [32..63]  - LI1:TS0-TS31 and LI5:TS0-TS31
        //    [64..95]  - LI2:TS0-TS31 and LI6:TS0-TS31
        //    [96..127] - LI3:TS0-TS31 and LI7:TS0-TS31
        //
        // upon connecting the pipe (see below), we have to pass the highway
        // number (0-1 - one highway per each QuadFalc (128TS per highway)
        // as well as the size of the highway (128 TS)
        //
        // this is just a simple algorithm to accomplish the above
        sqBurstTimeSlotMask[nTs + ((nLi % 4) * 32)] = 0xFF; 

        nCapacity += 64;
      }
    
      if (nCapacity != g_demoEnv.nRawDataCapacity) {
        DemoLogf("Error: Could not connect Raw Data Sender/Receiver with capacity=%d\n", g_demoEnv.nRawDataCapacity);
        OsaConsoleExit();
        OtxDrvReleaseLib();
        exit(1);
      }

    } else {
      // Subchannel
      nTs = 2;

      switch (g_demoEnv.nRawDataCapacity) {
      case 8:
        sqTimeSlotMask[nTs] = 0x01;
        break;
      case 16:
        sqTimeSlotMask[nTs] = 0x03;
        break;
      case 32:
        sqTimeSlotMask[nTs] = 0x0F;
        break;
      default:
        DemoLogf("Error: Unsupported capacity=%d\n", g_demoEnv.nRawDataCapacity);
        OsaConsoleExit();
        OtxDrvReleaseLib();
        exit(1);
        break;
      }

      // assign the same sub-channel
      sqBurstTimeSlotMask[nTs + ((nLi % 4) * 32)] = sqTimeSlotMask[nTs];

      nCapacity = g_demoEnv.nRawDataCapacity;
    }
  
    result = OtxDrvCreatePipe(nCapacity, &g_hPipes[nLi].hLiBurstRawData); DemoCheckRc(g_hPipes[nLi].hLiBurstRawData, result);
    result = OtxDrvCreatePipe(nCapacity, &g_hPipes[nLi].hBurstLiRawData); DemoCheckRc(g_hPipes[nLi].hBurstLiRawData, result);
  
    // Connect pipe: Burst -> Li 

    // Burst has 2 highways, 128TS per - so we have to specify which highway this pipe is connected
    result = OtxDrvConnectPipe(g_hPipes[nLi].hBurstLiRawData, OTX_PIPE_INPUT,  g_hPhysDevs.hBurst, iHighwayNo, sqBurstTimeSlotMask, cMaxHighwayTS);
    DemoCheckRc(g_hPipes[nLi].hBurstLiRawData, result);

    result = OtxDrvConnectPipe(g_hPipes[nLi].hBurstLiRawData, OTX_PIPE_OUTPUT, g_hPhysDevs.hLi[nLi],  -1, sqTimeSlotMask, 32);
    DemoCheckRc(g_hPipes[nLi].hBurstLiRawData, result);

    // Connect pipe: Li -> Burst 

    result = OtxDrvConnectPipe(g_hPipes[nLi].hLiBurstRawData, OTX_PIPE_INPUT,  g_hPhysDevs.hLi[nLi],  -1, sqTimeSlotMask, 32);
    DemoCheckRc(g_hPipes[nLi].hLiBurstRawData, result);

    // Burst has 2 highways, 128TS per - so we have to specify which highway this pipe is connected
    result = OtxDrvConnectPipe(g_hPipes[nLi].hLiBurstRawData, OTX_PIPE_OUTPUT, g_hPhysDevs.hBurst, iHighwayNo, sqBurstTimeSlotMask, cMaxHighwayTS);
    DemoCheckRc(g_hPipes[nLi].hLiBurstRawData, result);
  }

  // connect logical devices
  for (nLi = 0; nLi < g_demoEnv.coLi; nLi++) {
    // Receive direction:
    // Connect Raw Data Receiver
    result = OtxDrvConnectLogicalDevice(g_hPipes[nLi].hLiBurstRawData,  -1,  g_hLogicDevs.mRawDataReceiver[nLi].hHandle, -1);
    DemoCheckRc(g_hPipes[nLi].hLiBurstRawData, result);

    // Transmit direction:
    // Connect Raw Data Sender
    result = OtxDrvConnectLogicalDevice(g_hLogicDevs.mRawDataSender[nLi].hHandle, -1, g_hPipes[nLi].hBurstLiRawData, -1);
    DemoCheckRc(g_hLogicDevs.mRawDataSender[nLi].hHandle, result);
  }

}


void DemoInitializeLogicalDevices(
/*******************************************************************************

  @func
    Initialize the logical device. This function must be called after 
    the devices are created and connected.

*******************************************************************************/
  void
)
{
  OTX_RESULT result;
  OTX_INT32 i;

  for (i = 0; i < g_demoEnv.coLi; i++) {
    
    // Initialize the Raw Data Receivers (hosted in the DSP)
    result = OtxRawDataReceiverInit(
      g_hLogicDevs.mRawDataReceiver[i].hHandle, 
      OTX_DATA_FORMAT_UINT8,  // No data conversion. Deliver the data as it is received on the highway
      0xFF,        // Flag to be filtered
      OTX_FALSE,   // If OTX_TRUE, filter the specified flag
      OTX_TRUE     // Least Significant part is received first. Only applicable for a subchannel pipe (nCapacity < 64).
      ); DemoCheckRc(g_hLogicDevs.mRawDataReceiver[i].hHandle, result);

    // Initialize the Raw Data Senders (hosted in the DSP)
    result = OtxRawDataSenderInit(
      g_hLogicDevs.mRawDataSender[i].hHandle, 
      OTX_DATA_FORMAT_UINT8,  // No data conversion. 
      0xFF,        // Flag to be sent when the sender is out of data
      OTX_TRUE     // Least Significant part is transmitted first. Only applicable for a subchannel pipe (nCapacity < 64).
      ); DemoCheckRc(g_hLogicDevs.mRawDataSender[i].hHandle, result);
  }

}


void DemoEnableLogicalDevices(void)
/*******************************************************************************

  @func
    Enable Configured Logical Devices

*******************************************************************************/
{
  OTX_RESULT result;
  OTX_INT32 i;

  for (i = 0; i < g_demoEnv.coLi; i++) {

    DemoLogf("Enable mRawDataSender[%d]\n", i);
    result = OtxDrvEnable(g_hLogicDevs.mRawDataSender[i].hHandle, OTX_TASK_SYNC, 5000L); DemoCheckRc(g_hLogicDevs.mRawDataSender[i].hHandle, result);

#if 0 // We enable the Receiver in the DemoRawDataReceiver thread instead.
    OsaThreadSleep(100);
    DemoLogf("Enable mRawDataReceiver[%d]\n", i);
    result = OtxDrvEnable(g_hLogicDevs.mRawDataReceiver[i].hHandle, OTX_TASK_SYNC, 5000L); DemoCheckRc(g_hLogicDevs.mRawDataReceiver[i].hHandle, result);
#endif
  }

}


void DemoInitT1E1ConfigData(
/*******************************************************************************

  @func
    Initialize the T1/E1 Configuration Data Structures passed to the T1/E1
    Line Interface physical devices

*******************************************************************************/
  OUT OtxE1ConfigOptionsS *pE1CfgData,
  OUT OtxT1ConfigOptionsS *pT1CfgData
)
{
  // Initialize E1 Configuration Data
  pE1CfgData->eTxLineCode                = OTX_E1_LC_HDB3;
  pE1CfgData->eRxLineCode                = OTX_E1_LC_HDB3;
  pE1CfgData->eTxFrameFormat             = OTX_E1_FF_CRC4_MULTIFRAME;
  pE1CfgData->eRxFrameFormat             = OTX_E1_FF_CRC4_MULTIFRAME;
  pE1CfgData->eAisDetection              = OTX_E1_AIS_ETS300233;
  pE1CfgData->bfSiBits                   = 0x3;
  pE1CfgData->bfSaBits                   = 0x1F;
  pE1CfgData->bExtendedHDB3errorDetect   = OTX_FALSE;
  pE1CfgData->bAutoRegainMultiframe      = OTX_FALSE;
  pE1CfgData->bAutoResync                = OTX_TRUE;
  pE1CfgData->bAutoCrcBitInversion       = OTX_FALSE;
  pE1CfgData->bAutoRemoteAlarm           = OTX_FALSE;
  pE1CfgData->nLosSensitivity            = 128;
  pE1CfgData->nLosRecovery               = 128;
  pE1CfgData->nLineLength                = 10;
  pE1CfgData->bHighTxPower               = OTX_FALSE;
  pE1CfgData->bReceiveEqualizer          = OTX_FALSE;

  // Initialize T1 Configuration Data
  pT1CfgData->eSignalingMode             = OTX_T1_SM_CCS;
  pT1CfgData->eTxLineCode                = OTX_T1_LC_B8ZS;
  pT1CfgData->eRxLineCode                = OTX_T1_LC_B8ZS;
  pT1CfgData->eFrameFormat               = OTX_T1_FF_ESF;
  pT1CfgData->bCRC6Enable                = OTX_TRUE;
  pT1CfgData->eTxRemoteAlarmFormat       = OTX_T1_RA_YELLOW_A;
  pT1CfgData->eRxRemoteAlarmFormat       = OTX_T1_RA_YELLOW_A;
  pT1CfgData->bAutoResync                = OTX_TRUE;
  pT1CfgData->bAutoCrcBitInversion       = OTX_FALSE;
  pT1CfgData->eLfaSensitivity            = OTX_T1_LFA_2_OUT_OF_4;
  pT1CfgData->bAutoRemoteAlarm           = OTX_FALSE;
  pT1CfgData->nLosSensitivity            = 128;
  pT1CfgData->nLosRecovery               = 128;
  pT1CfgData->nLineLength                = 10;
  pT1CfgData->bHighTxPower               = OTX_FALSE;
  pT1CfgData->bReceiveEqualizer          = OTX_TRUE;
  pT1CfgData->bfClearChannels            = 0x00;
}


static void DemoCheckRc(IN OTX_HANDLE hDevice, IN OTX_RESULT result)
/*******************************************************************************
  @func
    Check the result from API call. If the result is a failure code then prompt
    the user if he wants to exit.
*******************************************************************************/
{
  if ( OTX_FAILURE(result) ) {

    if (hDevice == OTX_INVALID_HANDLE_VALUE) {
      DemoLogf("ERROR(%#lX) \n", result);
    } else {
      OTX_RESULT otxRc;
      char msgStr[256];
      OtxAttribTypeE  eAttribType;    // Type of the attribute, if attribute exists
      OTX_UINT32      cbAttribValue;  // Size in bytes of attribute value (Get value with <f OtxDrvGetAttributeValue>)
      OTX_BOOL        bReadOnly;      // Is attribute writable or read-only
      OTX_UINT32      eDeviceType = 999;    // Device type for hDevice 
      OTX_UINT32      nDeviceNumber = 999;  // Device number for hDevice
      
      // Get the Device Type for hHandle
      otxRc = OtxDrvGetAttributeType(hDevice, OTX_ATTR_DEVICE_TYPE, &eAttribType, &cbAttribValue, &bReadOnly); 
      if (OTX_SUCCESS(otxRc)) {
        assert(eAttribType == OTX_ATTRIB_UINT32);
        assert(cbAttribValue == sizeof(OTX_UINT32));
        OtxDrvGetAttributeValueUint32(hDevice, OTX_ATTR_DEVICE_TYPE, &eDeviceType); 
      } 
      
      // Get the Device Number for hHandle
      otxRc = OtxDrvGetAttributeType(hDevice, OTX_ATTR_DEVICE_NO, &eAttribType, &cbAttribValue, &bReadOnly); 
      if (OTX_SUCCESS(otxRc)) {
        assert(eAttribType == OTX_ATTRIB_UINT32);
        assert(cbAttribValue == sizeof(OTX_UINT32));
        OtxDrvGetAttributeValueUint32(hDevice, OTX_ATTR_DEVICE_NO, &nDeviceNumber); 
      }
      
      // Convert the result code to a more descriptive string
      otxRc = OtxDrvResultCode2String(hDevice, result, 256, &msgStr[0]);
      if (OTX_FAILURE(otxRc)) {
        strcpy(msgStr, "Cannot translate this result code.");
      }
      DemoLogf("ERROR(%#lX [%d.%d]) Dev(%X,%d,%d): %s \n", OTX_RESULT_CODE(result), OTX_RESULT_SOURCE_ID(result), OTX_RESULT_LINE_NUM(result), hDevice, (int) eDeviceType, (int)nDeviceNumber, msgStr);
    }

    DemoExitQuestion();
  }
}


/*******************************************************************************
  @func
    Prompts whether to exit the program. Usually called after an error 
    has been reported.

*******************************************************************************/
void DemoExitQuestion()
{
  short c;

  if (g_demoEnv.bExitOnError) {
    c = 'y';
  } else {
    DemoLogf("Do you want to exit (Y/N) ");
    c = OsaConsoleGetCh();
  }
  if (c == 'y' || c == 'Y') {
    // Clean up
    OtxDrvReleaseLib();
    OsaConsoleExit();
    exit(1);
  }
}


/*******************************************************************************
  @func
    Prints a timestamp in a MM-DD hh:mm:ss:mmm format, where

    MM  = month
    DD  = day (of month)
    hh  = hour of day
    mm  = minutes of day
    ss  = seconds of day
    mmm = milliseconds

*******************************************************************************/
void DemoPrintTimeStamp(
  OTX_DATETIME nTimeStamp
)
{
  OTX_DATETIME nTimeStampMs;
  OTX_DATETIME nTimeStampSec;
  struct Osa_tm *ptCurrentDayAndTime;
  Osa_time_t nCurrentTimeSecs;
  
  // Subtract number of 100ns ticks between Jan 1, 1970 and Jan 1, 1600
#if defined(__linux__)
  nTimeStamp.nUint64 = nTimeStamp.nUint64 - 1164447360 * 100000000;  
#elif defined(WIN32)
  nTimeStamp.nUint64 = nTimeStamp.nUint64 - 116444736000000000;  
#endif

  // Convert to seconds since Jan 1, 1970 (so we can use this time with OsaStdLocalTime())
  nTimeStampSec.nUint64 = nTimeStamp.nUint64 / 10000000;
  
  // Calculate number of milliseconds (since number of seconds is not a very accurate way to report timestamp)
  nTimeStampMs.nUint64 = nTimeStamp.nUint64 / 10000 - (nTimeStampSec.nUint64 * 1000);
  
  nCurrentTimeSecs = nTimeStampSec.sUint32.nLow;
  ptCurrentDayAndTime = OsaStdLocalTime(&nCurrentTimeSecs);
  
  if (ptCurrentDayAndTime) {
    
    DemoLogf("%.2d-%.2d %.2d:%.2d:%.2d.%.3d ", 
      ptCurrentDayAndTime->tm_mon + 1,
      ptCurrentDayAndTime->tm_mday,
      ptCurrentDayAndTime->tm_hour,
      ptCurrentDayAndTime->tm_min, 
      ptCurrentDayAndTime->tm_sec,
      nTimeStampMs.sUint32.nLow
      );
  } else {
    DemoLogf("Error: Invalid timestamp (before Jan 1, 1970)\n");
  }
    
}


/*******************************************************************************
  @func
    Idle Function that check for Events from the OTX Driver
*******************************************************************************/
void *DemoIdleFunction(
  void *lpParameter   // thread data
)
{
  OtxEventDataS eventData;
  OTX_RESULT result;

  while (1) {
    OtxDrvWaitForSingleEvent(g_hNotificationEvent, OTX_TIME_INFINITY);

    do {
      result = OtxDrvGetEventData(g_hEventQueue, &eventData); DemoCheckRc(g_hEventQueue, result);
      if (OTX_RESULT_CODE(result) == OTX_S_OK) {
        // We have a driver event
        // Printf the event data.
        // In a real application branch based on the data
        DemoPrintEvent(&eventData);
      }
    } while (OTX_RESULT_CODE(result) == OTX_S_OK);

  }

  return 0;
}


/*******************************************************************************
  @func
    Print DMA statistics.

    These stats are obtained from the driver, and are being used during
    initial release as a debugging feature.  These stats may be eliminated
    in the future.
*******************************************************************************/
void *DemoStatsThread(
  void *lpParameter   // thread data
)
{
  OTX_RESULT        result;

  OTX_UINT32        stat_Starts;          // Number of actual dma transfers initiated
  OTX_UINT32        stat_Completes;       // Number of actual dma transfers completed
  OTX_UINT32        stat_Partial;         // Number of partial transfer due to framemask
  OTX_UINT32        stat_Full;            // Number of count of max dma size transfers
  OTX_UINT32        stat_Aborts;          // Number of actual dma transfers aborts
  OTX_UINT32        stat_Amount;          // Number of amount of data transfered
  OTX_UINT32        stat_Errors;          // Number of count of processing errors
  OTX_UINT32        stat_OUruns;          // Number of amount of over/under runs
  OTX_UINT32        stat_DmaBusy;         // Number of count of DMA busy occurances
  OTX_UINT32        stat_Notify;          // Number of logical device notifies
  OTX_UINT32        stat_NotifyPartial;   // Number of logical device notifies - partial

  while (g_bStatsThreadActive) {

    result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hBurst, OTX_ATTR_BURST_STAT_RECV_STARTS,         &stat_Starts        ); DemoCheckRc(g_hPhysDevs.hBurst, result);
    result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hBurst, OTX_ATTR_BURST_STAT_RECV_COMPLETES,      &stat_Completes     ); DemoCheckRc(g_hPhysDevs.hBurst, result);
    result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hBurst, OTX_ATTR_BURST_STAT_RECV_PARTIAL,        &stat_Partial       ); DemoCheckRc(g_hPhysDevs.hBurst, result);
    result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hBurst, OTX_ATTR_BURST_STAT_RECV_FULL,           &stat_Full          ); DemoCheckRc(g_hPhysDevs.hBurst, result);
    result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hBurst, OTX_ATTR_BURST_STAT_RECV_ABORTS,         &stat_Aborts        ); DemoCheckRc(g_hPhysDevs.hBurst, result);
    result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hBurst, OTX_ATTR_BURST_STAT_RECV_AMOUNT,         &stat_Amount        ); DemoCheckRc(g_hPhysDevs.hBurst, result);
    result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hBurst, OTX_ATTR_BURST_STAT_RECV_ERRORS,         &stat_Errors        ); DemoCheckRc(g_hPhysDevs.hBurst, result);
    result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hBurst, OTX_ATTR_BURST_STAT_RECV_OURUNS,         &stat_OUruns        ); DemoCheckRc(g_hPhysDevs.hBurst, result);
    result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hBurst, OTX_ATTR_BURST_STAT_RECV_DMABUSY,        &stat_DmaBusy       ); DemoCheckRc(g_hPhysDevs.hBurst, result);
    result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hBurst, OTX_ATTR_BURST_STAT_RECV_NOTIFY,         &stat_Notify        ); DemoCheckRc(g_hPhysDevs.hBurst, result);
    result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hBurst, OTX_ATTR_BURST_STAT_RECV_NOTIFY_PARTIAL, &stat_NotifyPartial ); DemoCheckRc(g_hPhysDevs.hBurst, result);

    DemoLogf("Burst Recv:: Starts:%d, Completes:%d, Partial:=%d, Full=%d, Aborts:%d, Amount:%x, Errors:%d, OUErrors:%d, DMABusy:%d, Notify:%d, NotifyPartial:%d\n",
        stat_Starts,
        stat_Completes,
        stat_Partial,
        stat_Full,
        stat_Aborts,
        stat_Amount,
        stat_Errors,
        stat_OUruns,
        stat_DmaBusy,
        stat_Notify,
        stat_NotifyPartial    );

    result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hBurst, OTX_ATTR_BURST_STAT_SEND_STARTS,         &stat_Starts        ); DemoCheckRc(g_hPhysDevs.hBurst, result);
    result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hBurst, OTX_ATTR_BURST_STAT_SEND_COMPLETES,      &stat_Completes     ); DemoCheckRc(g_hPhysDevs.hBurst, result);
    result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hBurst, OTX_ATTR_BURST_STAT_SEND_PARTIAL,        &stat_Partial       ); DemoCheckRc(g_hPhysDevs.hBurst, result);
    result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hBurst, OTX_ATTR_BURST_STAT_SEND_FULL,           &stat_Full          ); DemoCheckRc(g_hPhysDevs.hBurst, result);
    result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hBurst, OTX_ATTR_BURST_STAT_SEND_ABORTS,         &stat_Aborts        ); DemoCheckRc(g_hPhysDevs.hBurst, result);
    result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hBurst, OTX_ATTR_BURST_STAT_SEND_AMOUNT,         &stat_Amount        ); DemoCheckRc(g_hPhysDevs.hBurst, result);
    result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hBurst, OTX_ATTR_BURST_STAT_SEND_ERRORS,         &stat_Errors        ); DemoCheckRc(g_hPhysDevs.hBurst, result);
    result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hBurst, OTX_ATTR_BURST_STAT_SEND_OURUNS,         &stat_OUruns        ); DemoCheckRc(g_hPhysDevs.hBurst, result);
    result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hBurst, OTX_ATTR_BURST_STAT_SEND_DMABUSY,        &stat_DmaBusy       ); DemoCheckRc(g_hPhysDevs.hBurst, result);
    result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hBurst, OTX_ATTR_BURST_STAT_SEND_NOTIFY,         &stat_Notify        ); DemoCheckRc(g_hPhysDevs.hBurst, result);
    result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hBurst, OTX_ATTR_BURST_STAT_SEND_NOTIFY_PARTIAL, &stat_NotifyPartial ); DemoCheckRc(g_hPhysDevs.hBurst, result);

    DemoLogf("Burst Send:: Starts:%d, Completes:%d, Partial:=%d, Full=%d, Aborts:%d, Amount:%x, Errors:%d, OUErrors:%d, DMABusy:%d, Notify:%d, NotifyPartial:%d\n",
        stat_Starts,
        stat_Completes,
        stat_Partial,
        stat_Full,
        stat_Aborts,
        stat_Amount,
        stat_Errors,
        stat_OUruns,
        stat_DmaBusy,
        stat_Notify,
        stat_NotifyPartial
    );

    OsaThreadSleep(2000); 
  }

  DemoLogf("DemoStatsThread() terminating.\n");

  return 0;
}


/*******************************************************************************
  @func
    Raw Data Receiver Thread Function 
*******************************************************************************/
void *DemoRawDataReceiver(
  void *lpParameter   // thread data
)
{
#define DEMO_RX_BUF_SIZE 8192
  OtxEventDataS eventData;
  OTX_RESULT result;
  OTX_RESULT otxRc;
  OTX_UINT8  buf[DEMO_RX_BUF_SIZE];
  OTX_UINT32 nFrameLen;
  OTX_BOOL bFrameEnd;
  char msgStr[256];
  clock_t startTime, stopTime;
  OTX_UINT32 nSourceId;
  OTX_UINT32 nWaitTime;

  startTime = clock();

  nSourceId = *((OTX_INT32 *) lpParameter);
  if (nSourceId < 0 || nSourceId > 7) {
    DemoLogf("DemoRawDataReceiver: Thread was passed an invalid nSourceId(%d). Thread exiting...\n", nSourceId);
    goto FATALEND;
  }

  g_hLogicDevs.mRawDataReceiver[nSourceId].bVerbose = OTX_FALSE;
  g_hLogicDevs.mRawDataReceiver[nSourceId].bActive  = OTX_TRUE;

  // Calculate (sleeping) time between each OtxRawDataReadData() call. (256ms fills 2048 bytes in a 64kbit/s pipe)
  nWaitTime = ((256 * 64) / g_demoEnv.nRawDataCapacity) * 2; 

  // Enable the receiver
  result = OtxDrvEnable(g_hLogicDevs.mRawDataReceiver[nSourceId].hHandle, OTX_TASK_SYNC, 5000L); DemoCheckRc(g_hLogicDevs.mRawDataReceiver[nSourceId].hHandle, result);

  // Discard any events from event queue (if any)
  do {
    result = OtxDrvGetEventData(g_hLogicDevs.mRawDataReceiver[nSourceId].hEventQueue, &eventData); DemoCheckRc(g_hLogicDevs.mRawDataReceiver[nSourceId].hEventQueue, result);
  } while (OTX_RESULT_CODE(result) == OTX_S_OK);

  // Discard any data from the Raw Data Receiver
  do {
    result = OtxRawDataReadData(g_hLogicDevs.mRawDataReceiver[nSourceId].hHandle, buf, DEMO_RX_BUF_SIZE, &nFrameLen, &bFrameEnd);
  } while (OTX_RESULT_CODE(result) == OTX_S_OK);
  if (OTX_RESULT_CODE(result) != OTX_S_FALSE) {
    DemoLogf("RawDataReceiver[%d] OtxRawDataReadData() was unsuccessful.\n", nSourceId);
    DemoCheckRc(g_hLogicDevs.mRawDataReceiver[nSourceId].hHandle, result);
  }

  DemoLogf("DemoRawDataReceiver[%d]: Waiting for events. Reading data every %d milliseconds.\n", nSourceId, nWaitTime);

WAIT_FOR_EVENTS:

  while (g_hLogicDevs.mRawDataReceiver[nSourceId].bActive) {

    // Wait for a specified time for any (unexpected) events from the Raw Data Receiver device
    OtxDrvWaitForSingleEvent(
      g_hLogicDevs.mRawDataReceiver[nSourceId].hNotificationEvent, 
      nWaitTime
    );
    //DemoLogf("RawDataReceiver[%d] Received Event\n");
    do {
      result = OtxDrvGetEventData(g_hLogicDevs.mRawDataReceiver[nSourceId].hEventQueue, &eventData); DemoCheckRc(g_hLogicDevs.mRawDataReceiver[nSourceId].hEventQueue, result);

      if (OTX_RESULT_CODE(result) == OTX_S_OK) {
        // We have a driver event
        switch (eventData.m_eDeviceType) {
        case OTX_LDEVICE_DATA_RAW_RECEIVER:
          // Event from Raw Data Receiver 
          if ((eventData.m_nSourceId != nSourceId)) {
            DemoLogf("RawDataReceiver[%d] Error: m_nSourceId is out of range.\n", eventData.m_nSourceId);
            break;
          } 

          switch (eventData.m_nCode) {
          case OTX_RAWDATA_RECEIVER_EC_DATA_AVAILABLE:
            DemoLogf("RawDataReceiver[%d] Unexpected. Received OTX_RAWDATA_RECEIVER_EC_DATA_AVAILABLE event.\n", eventData.m_nSourceId);
            break;
          default:
            otxRc = OtxDrvEventCode2String(eventData.m_hDevice, eventData.m_nCode, 256, &msgStr[0]);
            // Unexpected event from this device
            if (OTX_SUCCESS(otxRc)) {
              DemoLogf("RawDataReceiver[%d] Event: %s \n", eventData.m_nSourceId, msgStr);
            } else {
              DemoLogf("RawDataReceiver[%d] Invalid Event: %ld \n", eventData.m_nSourceId, eventData.m_nCode);
            }
            break;
          }
          break;

          default:
            // Unhandled event
            DemoLogf("DemoRawDataReceiver[%d]: Event from an unknown object 0x%X\n", nSourceId, eventData.m_hDevice);
            break;
        }
        

      }
    } while (OTX_RESULT_CODE(result) == OTX_S_OK);

    // Read data from the device
    otxRc = OtxRawDataReadData(
      g_hLogicDevs.mRawDataReceiver[nSourceId].hHandle, 
      buf, 
      DEMO_RX_BUF_SIZE, 
      &nFrameLen, 
      &bFrameEnd);
    while (OTX_RESULT_CODE(otxRc) == OTX_S_OK) {
      OTX_INT32 x;
      
      if (g_hLogicDevs.mRawDataReceiver[nSourceId].bVerbose) {
        DemoLogf("RawDataReceiver[%d] (len=%.4d) ", nSourceId, nFrameLen);
        for (x = 0; (x < 16) && (x < (OTX_INT32)nFrameLen); x++) {
          DemoLogf("%.2X ", buf[x]);
        }
        DemoLogf("\n");
      }
      
      if (g_hLogicDevs.mRawDataReceiver[nSourceId].pRxFile) {
        int cbWritten;
        
        cbWritten = fwrite(buf, sizeof(OTX_UINT8), nFrameLen, g_hLogicDevs.mRawDataReceiver[nSourceId].pRxFile);
        
        if (cbWritten != (int)nFrameLen) {
          DemoLogf("RawDataReceiver[%d] Error: Could not write data to file.  (nFrameLen=%d, cbWritten=%d) \n", eventData.m_nSourceId, nFrameLen, cbWritten);
        }
        
        if (bFrameEnd) {
          DemoLogf("RawDataReceiver[%d] Data reception complete. Closing data file. \n", nSourceId);
          fclose(g_hLogicDevs.mRawDataReceiver[nSourceId].pRxFile);
          g_hLogicDevs.mRawDataReceiver[nSourceId].pRxFile = NULL;
          goto END;
        }
      }

      if (nFrameLen != DEMO_RX_BUF_SIZE) {
        // No more data to read at this time. Break this while loop
        goto WAIT_FOR_EVENTS;
      }
      
      DemoLogf("RawDataReceiver[%d] Reading more data. [Abnormal situation]\n", nSourceId);
      otxRc = OtxRawDataReadData(
        g_hLogicDevs.mRawDataReceiver[nSourceId].hHandle, 
        buf, 
        DEMO_RX_BUF_SIZE, 
        &nFrameLen, 
        &bFrameEnd);
    }
    if (OTX_RESULT_CODE(otxRc) != OTX_S_FALSE) {
      DemoLogf("RawDataReceiver[%d] OtxRawDataReadData() was unsuccessful.\n", nSourceId);
      DemoCheckRc(g_hLogicDevs.mRawDataReceiver[nSourceId].hHandle, otxRc);
    }
    
  }

END:
  
  stopTime = clock();

  // Disable the receiver
  result = OtxDrvDisable(g_hLogicDevs.mRawDataReceiver[nSourceId].hHandle, OTX_TASK_SYNC, 5000L); DemoCheckRc(g_hLogicDevs.mRawDataReceiver[nSourceId].hHandle, result);
  
  DemoLogf("DemoRawDataReceiver[%d]: Done (time=%2.1f seconds).\n", nSourceId, (double)(stopTime - startTime) / CLOCKS_PER_SEC);

FATALEND:
  return 0;

}

/*******************************************************************************
  @func
    Raw Data Sender Thread Function 
*******************************************************************************/
void *DemoRawDataSender(
  void *lpParameter   // thread data
)
{
  OtxEventDataS eventData;
  OTX_RESULT result;
  OTX_RESULT otxRc;
  char msgStr[256];
  clock_t startTime, stopTime;
  OTX_UINT32 nSourceId;
  OTX_UINT32 nTaskRef = 0;
  OTX_INT32 cbPrimed = 0;
  

  startTime = clock();

  nSourceId = *((OTX_INT32 *) lpParameter);
  if (nSourceId < 0 || nSourceId > 7) {
    DemoLogf("DemoRawDataSender: Thread was passed an invalid nSourceId(%d). Thread exiting...\n", nSourceId);
    goto END;
  }

  g_hLogicDevs.mRawDataSender[nSourceId].bActive = OTX_TRUE;
  g_hLogicDevs.mRawDataSender[nSourceId].coUnderRunErrors = 0;

  // Discard any events from event queue (if any)
  do {
    result = OtxDrvGetEventData(g_hLogicDevs.mRawDataSender[nSourceId].hEventQueue, &eventData); DemoCheckRc(g_hLogicDevs.mRawDataSender[nSourceId].hEventQueue, result);
  } while (OTX_RESULT_CODE(result) == OTX_S_OK);

  if (g_hLogicDevs.mRawDataSender[nSourceId].pTxFile) {
    // Send from a file
    do {
      g_hLogicDevs.mRawDataSender[nSourceId].cbTxBuf = fread(g_hLogicDevs.mRawDataSender[nSourceId].sqTxBuf, sizeof(OTX_UINT8), DEMO_RAW_TX_BUF_SIZE, g_hLogicDevs.mRawDataSender[nSourceId].pTxFile);
      
      otxRc = OtxRawDataSendData(g_hLogicDevs.mRawDataSender[nSourceId].hHandle, g_hLogicDevs.mRawDataSender[nSourceId].sqTxBuf, g_hLogicDevs.mRawDataSender[nSourceId].cbTxBuf, nTaskRef, OTX_TIME_INFINITY);
      if (OTX_RESULT_CODE(otxRc) != OTX_E_BUSY) {
        DemoCheckRc(g_hLogicDevs.mRawDataSender[nSourceId].hHandle, otxRc);
        DemoLogf("DemoRawDataSender[%d]: Packet #%d buffered\n", nSourceId, nTaskRef);
        nTaskRef++;
      }
      
    } while (g_hLogicDevs.mRawDataSender[nSourceId].cbTxBuf && (OTX_RESULT_CODE(otxRc) != OTX_E_BUSY));
    
    if (g_hLogicDevs.mRawDataSender[nSourceId].cbTxBuf) {
      //DemoLogf("DemoRawDataSender[%d]: Waiting for OTX_RAWDATA_SENDER_EC_BUF_DEPLETED events to complete the transmission...\n", nSourceId);
    } else {
      DemoLogf("DemoRawDataSender[%d]: Transmission complete. Closing file\n", nSourceId);
      fclose(g_hLogicDevs.mRawDataSender[nSourceId].pTxFile);
      g_hLogicDevs.mRawDataSender[nSourceId].pTxFile = NULL;
      goto END;
    }
  } else {
    // Send a text string (continously until aborted)
    OTX_INT32 cbText = strlen(DEMO_TEST_STRING);
    OTX_INT32 j;

    DemoLogf("DemoRawDataSender[%d]: Preparing text string buffer...\n", nSourceId);
    
    g_hLogicDevs.mRawDataSender[nSourceId].sqTxBuf[0] = '\0';

    // Fill sqTxBuf with the text string (as many copies of the text string that will fit in sqTxBuf). 
    for (j = 0; j < (DEMO_RAW_TX_BUF_SIZE-1)/cbText; j++) {
      strcpy((char *)g_hLogicDevs.mRawDataSender[nSourceId].sqTxBuf + strlen((char *)g_hLogicDevs.mRawDataSender[nSourceId].sqTxBuf), DEMO_TEST_STRING);
    }

    g_hLogicDevs.mRawDataSender[nSourceId].cbTxBuf = strlen((char *)g_hLogicDevs.mRawDataSender[nSourceId].sqTxBuf);

    DemoLogf("DemoRawDataSender[%d]: Done preparing text string buffer (len=%d)...\n", nSourceId, g_hLogicDevs.mRawDataSender[nSourceId].cbTxBuf);

    // Prime the sender
    cbPrimed = 0;
    do {
      otxRc = OtxRawDataSendData(g_hLogicDevs.mRawDataSender[nSourceId].hHandle, g_hLogicDevs.mRawDataSender[nSourceId].sqTxBuf, g_hLogicDevs.mRawDataSender[nSourceId].cbTxBuf, nTaskRef, OTX_TIME_INFINITY);
      if (OTX_RESULT_CODE(otxRc) != OTX_E_BUSY) {
        DemoCheckRc(g_hLogicDevs.mRawDataSender[nSourceId].hHandle, otxRc);
        cbPrimed += g_hLogicDevs.mRawDataSender[nSourceId].cbTxBuf;
        //DemoLogf("DemoRawDataSender[%d]: Packet #%d buffered\n", nSourceId, nTaskRef);
        nTaskRef++;
      }

    } while (OTX_RESULT_CODE(otxRc) != OTX_E_BUSY);
    DemoLogf("DemoRawDataSender[%d]: Primed %d bytes\n", nSourceId, cbPrimed);
     
  }
  
  
  while (g_hLogicDevs.mRawDataSender[nSourceId].bActive) {
    otxRc = OtxDrvWaitForSingleEvent(g_hLogicDevs.mRawDataSender[nSourceId].hNotificationEvent, 500);
    if (OTX_RESULT_CODE(otxRc) == OTX_S_TIMEOUT) {
      continue;   // allow thread abort signal a chance
    }
    do {
      result = OtxDrvGetEventData(g_hLogicDevs.mRawDataSender[nSourceId].hEventQueue, &eventData); DemoCheckRc(g_hLogicDevs.mRawDataSender[nSourceId].hEventQueue, result);
      
      if (OTX_RESULT_CODE(result) == OTX_S_OK) {
        // We have a driver event
        switch (eventData.m_eDeviceType) {
        case OTX_LDEVICE_DATA_RAW_SENDER:
          // Event from Raw Data Sender 
          if ((eventData.m_nSourceId != nSourceId)) {
            DemoLogf("RawDataSender[%d] Error: m_nSourceId is out of range.\n", eventData.m_nSourceId);
            break;
          } 
          
          switch (eventData.m_nCode) {
          case OTX_RAWDATA_SENDER_EC_COMPLETED:
            if (!g_hLogicDevs.mRawDataSender[nSourceId].pTxFile) {
              g_hLogicDevs.mRawDataSender[nSourceId].coUnderRunErrors++;
              DemoLogf("DemoRawDataSender[%d]: *** Data Underrun (m_nParam=%d).\n", eventData.m_nSourceId, eventData.m_nParam);
              // Fall-through
            } else {
              DemoLogf("DemoRawDataSender[%d]: Transmission ended (no more data to be transmitted).\n", eventData.m_nSourceId);
              fclose(g_hLogicDevs.mRawDataSender[nSourceId].pTxFile);
              g_hLogicDevs.mRawDataSender[nSourceId].pTxFile = NULL;
              goto END;
            }
            // Fall-through
          case OTX_RAWDATA_SENDER_EC_BUF_DEPLETED:
            {
              OTX_RESULT otxRc;
              
              if (g_hLogicDevs.mRawDataSender[nSourceId].pTxFile) {
                // Sending from file
                DemoLogf("DemoRawDataSender[%d]: Packet #%d sent\n", eventData.m_nSourceId, eventData.m_nRequestId);
                
                if (g_hLogicDevs.mRawDataSender[nSourceId].cbTxBuf) {
                  
                  while (g_hLogicDevs.mRawDataSender[nSourceId].cbTxBuf) {
                    otxRc = OtxRawDataSendData(g_hLogicDevs.mRawDataSender[nSourceId].hHandle, g_hLogicDevs.mRawDataSender[nSourceId].sqTxBuf, g_hLogicDevs.mRawDataSender[nSourceId].cbTxBuf, nTaskRef, OTX_TIME_INFINITY);
                    if (OTX_RESULT_CODE(otxRc) == OTX_E_BUSY) {
                      OsaThreadSleep((1024*64)/g_demoEnv.nRawDataCapacity); // Give the driver some time to send the data (1024ms sends 8192 bytes in a 64kbit/s pipe)
                      break;
                    } else {
                      DemoCheckRc(g_hLogicDevs.mRawDataSender[nSourceId].hHandle, otxRc);
                    }
                    
                    DemoLogf("DemoRawDataSender[%d]: Packet #%d buffered\n", nSourceId, nTaskRef);
                    nTaskRef++;
                    
                    g_hLogicDevs.mRawDataSender[nSourceId].cbTxBuf = fread(g_hLogicDevs.mRawDataSender[nSourceId].sqTxBuf, sizeof(OTX_UINT8), DEMO_RAW_TX_BUF_SIZE, g_hLogicDevs.mRawDataSender[nSourceId].pTxFile);
                  }
                  
                  if (!g_hLogicDevs.mRawDataSender[nSourceId].cbTxBuf) {
                    DemoLogf("DemoRawDataSender[%d]: EOF reached. Waiting for completion.\n", eventData.m_nSourceId);
                  }
                } else {
                  DemoLogf("DemoRawDataSender[%d]: Waiting for completion.\n", eventData.m_nSourceId);
                }
              } else if (g_hLogicDevs.mRawDataSender[nSourceId].cbTxBuf) {
                // Sending from static buffer (continously until aborted)
                while (g_hLogicDevs.mRawDataSender[nSourceId].bActive) {
                  otxRc = OtxRawDataSendData(g_hLogicDevs.mRawDataSender[nSourceId].hHandle, g_hLogicDevs.mRawDataSender[nSourceId].sqTxBuf, g_hLogicDevs.mRawDataSender[nSourceId].cbTxBuf, nTaskRef, OTX_TIME_INFINITY);
                  if (OTX_RESULT_CODE(otxRc) == OTX_E_BUSY) {
                    OsaThreadSleep((1024*64)/g_demoEnv.nRawDataCapacity); // Give the driver some time to send the data (1024ms sends 8192 bytes in a 64kbit/s pipe)
                    break;
                  } else {
                    DemoCheckRc(g_hLogicDevs.mRawDataSender[nSourceId].hHandle, otxRc);
                  }
                  
                  //DemoLogf("DemoRawDataSender[%d]: Packet #%d buffered\n", nSourceId, nTaskRef);
                  nTaskRef++;
                }
              } else {
                // No file handles and no more bytes to send. Waiting for the OTX_RAWDATA_SENDER_EC_COMPLETED event
              }
            }
            break;
            
            
          default:
            otxRc = OtxDrvEventCode2String(eventData.m_hDevice, eventData.m_nCode, 256, &msgStr[0]);
            // Unexpected event from this device
            if (OTX_SUCCESS(otxRc)) {
              DemoLogf("RawDataSender[%d] Event: %s \n", eventData.m_nSourceId, msgStr);
            } else {
              DemoLogf("RawDataSender[%d] Invalid Event: %ld \n", eventData.m_nSourceId, eventData.m_nCode);
            }
            break;
          }
          break;
          default:
            // Unhandled event
            DemoLogf("DemoRawDataSender[%d]: Event from an unknown object 0x%X\n", nSourceId, eventData.m_hDevice);
            break;
        }
        
      }
    } while (OTX_RESULT_CODE(result) == OTX_S_OK);
  }

END:
  
  stopTime = clock();
  
  DemoLogf("DemoRawDataSender[%d]: Done (time=%2.1f seconds, coUnderRunErrors=%d).\n", nSourceId, (double)(stopTime - startTime) / CLOCKS_PER_SEC, g_hLogicDevs.mRawDataSender[nSourceId].coUnderRunErrors);
  
  return 0;
}


void DemoPrintEvent(
/*******************************************************************************
  @func
    Prints a Driver Event. 
*******************************************************************************/
  OtxEventDataS *pEventData
)
{
  OTX_RESULT otxRc;
  char msgStr[256];

  //EnterCriticalSection(&g_analyzeCritSec);


  DemoPrintTimeStamp(pEventData->m_nTimeStamp);
    
  switch (pEventData->m_eDeviceType) {
    case OTX_DEVICE_LI_T1E1:
    // Event from Li0 or Li1
    otxRc = OtxDrvEventCode2String(pEventData->m_hDevice, pEventData->m_nCode, 256, &msgStr[0]);
    if (OTX_SUCCESS(otxRc)) {
      if (pEventData->m_nParam) {
        DemoLogf("E1/T1[%d] Event: %s (%d)\n", pEventData->m_nSourceId, msgStr, pEventData->m_nParam);
      } else {
        DemoLogf("E1/T1[%d] Event: %s\n", pEventData->m_nSourceId, msgStr);
      }
    } else {
      DemoLogf("E1/T1[%d] Invalid Event: %ld \n", pEventData->m_nSourceId, pEventData->m_nCode);
    }
    break;

  case OTX_DEVICE_BURST:
    otxRc = OtxDrvEventCode2String(pEventData->m_hDevice, pEventData->m_nCode, 256, &msgStr[0]);
    if (OTX_SUCCESS(otxRc)) {
      DemoLogf("Burst[%d] Event: %s \n", pEventData->m_nSourceId, msgStr);
    } else {
      DemoLogf("Burst[%d] Invalid Event: %ld \n", pEventData->m_nSourceId, pEventData->m_nCode);
    }
    break;
    
  default:
    // Unhandled event
    DemoLogf("Event from an unknown object 0x%X\n", pEventData->m_hDevice);
    break;
  }


  //LeaveCriticalSection(&g_analyzeCritSec);

}

OTX_INT32 DemoProcessUserCommand(char commandChar)
/*******************************************************************************

  @func

  Processes User Commands

  @comm

*******************************************************************************/
{
  static OTX_INT32 nCurrentLi = 0;
  OTX_RESULT result;

  DemoLogf("Key '%c' pressed\n", commandChar);

  switch (commandChar) {

  case '?':
  case 'h':
  case 'H':
    DemoLogf("Cmd  Synopsis\n");
    DemoLogf("===  ========\n");
    DemoLogf(" ?   Display Help text\n");
    DemoLogf(" h   Ditto\n");
    DemoLogf(" q   Quit\n");
    DemoLogf(" 0-7 Select active Li or Dsp, this Li/Dsp is then used for subsequent commands\n");
    DemoLogf(" x   Send a text file using the RawData Sender[%d]. Received by the RawData Receiver[%d]\n", nCurrentLi, nCurrentLi + 1);
    DemoLogf(" a   Abort all Raw Data Senders and Raw Data Receivers\n");
    DemoLogf(" r   Start Raw Data Receiver[%d] (no file logging)\n", nCurrentLi);
    DemoLogf(" R   Start Raw Data Receiver[%d] Prompt for file to recv\n", nCurrentLi);
    DemoLogf(" s   Start Raw Data Sender[%d]. Continously send a text string\n", nCurrentLi);
    DemoLogf(" S   Start Raw Data Sender[%d]. Prompt for file to send\n", nCurrentLi);
    DemoLogf(" b   Start Burst Sender[%d]. Continously send a text string\n", nCurrentLi / 4);
    DemoLogf(" B   Start Burst Sender[%d]. Prompt for file to send\n", nCurrentLi / 4);
    DemoLogf(" v   Toggle verbose (print received data) setting Raw Data Receiver[%d]\n", nCurrentLi);
    DemoLogf(" l   Query Line Status of both Lis\n");
    DemoLogf(" t   Toggle verbose buffer usage statistics\n");
    DemoLogf(" f   Forcing resynchronization\n");
    DemoLogf(" F   Reset all lines\n");
    DemoLogf(" T   Reset buffer usage statistics\n");
    DemoLogf(" D   Dump buffer usage statistics\n");
    break;

  case 't':
    g_bStatsThreadActive = !g_bStatsThreadActive;

    if (g_bStatsThreadActive) {
      DemoLogf("Starting Stats Thread\n");
      OsaThreadStart(DemoStatsThread);
    }
    break;

  case 'f':
    {
      OTX_INT32 nLi;

      for (nLi = 0; nLi < g_demoEnv.coLi; nLi++) {
        DemoLogf("Forcing resynchronization on Li #%d... ", nLi);
        result = OtxT1E1ForceResynch(g_hPhysDevs.hLi[nLi]); DemoCheckRc(g_hPhysDevs.hLi[nLi], result);
        DemoLogf("Done.");
      }
    }
    break;

  case 'F':
    {
      OTX_INT32 nLi;

      for (nLi = 0; nLi < g_demoEnv.coLi; nLi++) {
        DemoLogf("Reseting Li #%d... ", nLi);
        result = OtxDrvReset(g_hPhysDevs.hLi[nLi]); DemoCheckRc(g_hPhysDevs.hLi[nLi], result);
        DemoLogf("Done.");
      }
    }
    break;

  case 'x':
    {
      char sqzFileName[64];

      DemoLogf("RawDataSender[%d] => RawDataReceiver[%d]. Please enter file to send: ", nCurrentLi, nCurrentLi + 1);
      OsaConsoleGetStr(sqzFileName);

      if (sqzFileName[0] == '\0') {
        break;
      }
      if ((g_hLogicDevs.mRawDataSender[nCurrentLi].pTxFile = fopen(sqzFileName, "rb")) == NULL) {
        DemoLogf("Error: Could not open file %s \n", sqzFileName);
        break;
      }

      if ((g_hLogicDevs.mRawDataReceiver[nCurrentLi + 1].pRxFile = fopen("T2mRawData.Raw", "wb")) == NULL) {
        DemoLogf("Error: Could not open file %s \n", sqzFileName);
        break;
      }

      DemoLogf("Creating Raw Data Receiver[%d] Thread\n", nCurrentLi + 1);
    	OsaThreadStartEx(DemoRawDataReceiver, (void *)&g_hLogicDevs.mRawDataReceiver[nCurrentLi + 1].nSourceId);

      DemoLogf("Creating Raw Data Sender[%d] Thread\n", nCurrentLi);
    	OsaThreadStartEx(DemoRawDataSender, (void *)&g_hLogicDevs.mRawDataSender[nCurrentLi].nSourceId);
    }
    break;

  case 'a':
    g_hLogicDevs.mRawDataReceiver[nCurrentLi].bActive   = OTX_FALSE;
    g_hLogicDevs.mRawDataSender[nCurrentLi].bActive     = OTX_FALSE;
    break;

  case 'v':
    g_hLogicDevs.mRawDataReceiver[nCurrentLi].bVerbose = !g_hLogicDevs.mRawDataReceiver[nCurrentLi].bVerbose;
    DemoLogf("Setting verbose setting to %s for Raw Data Receiver[%d] Thread\n", (g_hLogicDevs.mRawDataReceiver[nCurrentLi].bVerbose)?"ON":"OFF", nCurrentLi);
    break;

  case 'r':
    if (g_hLogicDevs.mRawDataReceiver[nCurrentLi].bActive == OTX_FALSE) {
      g_hLogicDevs.mRawDataReceiver[nCurrentLi].pRxFile = NULL;
      DemoLogf("Creating Raw Data Receiver[%d] Thread\n", nCurrentLi);
      OsaThreadStartEx(DemoRawDataReceiver, (void *)&g_hLogicDevs.mRawDataReceiver[nCurrentLi].nSourceId);
    } else {
      DemoLogf("Raw Data Receiver[%d] Thread already running.\n", nCurrentLi);
    }
    break;

  case 'R':
    {
      char sqzFileName[OTX_MAX_PATH];

      if (g_hLogicDevs.mRawDataReceiver[nCurrentLi].bActive == OTX_FALSE) {
        DemoLogf("RawDataReceiver[%d]. Please enter file name: ", nCurrentLi);
        OsaConsoleGetStr(sqzFileName);
        if (sqzFileName[0] == '\0') {
          break;
        }

        if ((g_hLogicDevs.mRawDataReceiver[nCurrentLi].pRxFile = fopen(sqzFileName, "wb")) == NULL) {
          DemoLogf("Error: Could not open file %s \n", sqzFileName);
          break;
        }

        DemoLogf("Creating Raw Data Receiver[%d] Thread\n", nCurrentLi);
        OsaThreadStartEx(DemoRawDataReceiver, (void *)&g_hLogicDevs.mRawDataReceiver[nCurrentLi].nSourceId);
      } else {
        DemoLogf("Raw Data Receiver[%d] Thread already running.\n", nCurrentLi);
      }
    }
    break;

  case 'S':
    {
      char sqzFileName[OTX_MAX_PATH];

      if (g_hLogicDevs.mRawDataSender[nCurrentLi].bActive == OTX_FALSE) {
        DemoLogf("RawDataSender[%d]. Please enter file to send: ", nCurrentLi);
        OsaConsoleGetStr(sqzFileName);
        if (sqzFileName[0] == '\0') {
          break;
        }

        if ((g_hLogicDevs.mRawDataSender[nCurrentLi].pTxFile = fopen(sqzFileName, "rb")) == NULL) {
          DemoLogf("Error: Could not open file %s \n", sqzFileName);
          break;
        }

        DemoLogf("Creating Raw Data Sender[%d] Thread\n", nCurrentLi);
    	  OsaThreadStartEx(DemoRawDataSender, (void *)&g_hLogicDevs.mRawDataSender[nCurrentLi].nSourceId);
      } else {
        DemoLogf("Raw Data Sender[%d] Thread already running.\n", nCurrentLi);
      }
    }
    break;

  case 's':
    {
      if (g_hLogicDevs.mRawDataSender[nCurrentLi].bActive == OTX_FALSE) {
        DemoLogf("RawDataSender[%d]. Continously send text string \"%s\".\n", nCurrentLi, DEMO_TEST_STRING);

        g_hLogicDevs.mRawDataSender[nCurrentLi].pTxFile = NULL;

        DemoLogf("Creating Raw Data Sender[%d] Thread\n", nCurrentLi);
      	OsaThreadStartEx(DemoRawDataSender, (void *)&g_hLogicDevs.mRawDataSender[nCurrentLi].nSourceId);
      } else {
        DemoLogf("Raw Data Sender[%d] Thread already running.\n", nCurrentLi);
      }
    }
    break;

  case 'l':
    {
      OtxT1E1Layer1StatusE eLayer1Status;
      OTX_INT32 nLi;

      for (nLi = 0; nLi < g_demoEnv.coLi; nLi++) {
        DemoLogf("Line status of Li #%d is ", nLi);
        result = OtxT1E1GetLayer1Status(g_hPhysDevs.hLi[nLi], &eLayer1Status); DemoCheckRc(g_hPhysDevs.hLi[nLi], result);
        DemoLogf("%s\n", eLayer1Status == OTX_T1E1_LAYER1_OK ? "OK" : "DOWN");
      }
    }
    break;

  case '0':
  case '1':
  case '2':
  case '3':
  case '4':
  case '5':
  case '6':
  case '7':
    // Select current Li
    nCurrentLi = commandChar - '0';
    DemoLogf("LI#%d Selected ....\n", nCurrentLi);
    break;

  case 'q':
  case 'Q':
    return 1;
    break;

  case 'T':
    // reset stats
    OtxDrvDbgDump(g_hPhysDevs.hBurst, 101);
    break;

  case 'D':
    // dump stats to debugger
    OtxDrvDbgDump(g_hPhysDevs.hBurst, 100);
    break;

  default:
    DemoLogf("Unknown command. Press '?' for help, 'q' to quit\n");
    break;
  }

  return 0;
}                      


/*******************************************************************************

  @func

  Parse the program's command line arguments and calls upon
  requested actions.

  There are some optional switches that can be used following
  the driver name on the command line. Switches are are
  separated by spaces, and can appear in arbitrary order.
  Switches consist of a hyphen and a letter:

  @comm

*******************************************************************************/
static void DemoParseCmdLineArgs(int argc, char *argv[], DemoEnvT *envP)
{
  short  i;

  // Set reasonable defaults before reading from command line
  envP->pszInputStr       = NULL;
  envP->bExitOnError      = OTX_FALSE;
  envP->eT1E1Mode         = OTX_T1E1_MODE_E1;
  envP->boardNo           = DEMO_DEFAULT_BOARD_NO;
  envP->nBoardType        = OTX_DEVICE_THOR_8_PCI_PLUS;
  envP->coLi              = 8;
  envP->nRawDataCapacity  = 64;
  envP->nStartingTimeslot = 1;

  strcpy(envP->logFileName, envP->cwd);
  strcat(envP->logFileName, DEMO_DEFAULT_LOG_FILE);
  for (i = 1; i < argc; i++) {        // Scan command line arguments
    if (*(argv[i]+1) == 'Z') {   // Command input string -- case sensitive
      envP->pszInputStr = argv[i]+2;
    } else {
      OsaStdStrUpr(argv[i]);
      if (*argv[i] == '-' || *argv[i] == '/') {  // Is argument a legal switch?
        if (*(argv[i]+1) == 'L') {   // Log file
          strcpy(envP->logFileName, argv[i] + 2);
        } else if (*(argv[i]+1) == 'B') {   // Board number
          envP->boardNo = strtoul(argv[i]+2, NULL, 0);
        } else if (*(argv[i]+1) == 'T') {   // Board Type
          if ( *(argv[i]+2) == 'P') {
            envP->nBoardType = OTX_DEVICE_THOR_8_PCI_PLUS;
            envP->coLi  = 8;
          } else if ( *(argv[i]+2) == 'S') {
            envP->nBoardType = OTX_DEVICE_SLEIPNIR_1_PCI;
            envP->coLi  = 1;
          } else if (*(argv[i]+1) == 'Q') {   // Exit on error
            envP->bExitOnError = OTX_TRUE;
          } else {
            DemoLogf("Error: Unknown command line switch %s\n", argv[i]);
            DemoHelpPanel();
            OsaConsoleExit();
            exit(1);
          }
        
        } else if (*(argv[i]+1) == 'S') {   // Starting timeslot
          envP->nStartingTimeslot = strtol(argv[i]+2, NULL, 0);

        } else if (*(argv[i]+1) == 'N') {   // Capacity of Raw Data Pipe
          envP->nRawDataCapacity = strtol(argv[i]+2, NULL, 0);
          if (envP->nRawDataCapacity >= 64) {
            if ((envP->nRawDataCapacity%64) != 0) {
              DemoLogf("Error: Raw Data Pipe capacity must be a multiple of 64, or a subchannel capacity (8, 16, or 32)\n");
              DemoHelpPanel();
              OsaConsoleExit();
              exit(1);
            }
          } else {
            switch (envP->nRawDataCapacity) {
            case 8:
            case 16:
            case 32:
              DemoLogf("When using subchannels with this demo program, please be aware the output\n");
              DemoLogf("file produced by this program might not be identical to the input\n");
              DemoLogf("file, because the transmitted octets might not be bit-aligned with the\n");
              DemoLogf("received octets. In order to compare the output file with the input file, \n");
              DemoLogf("the output file might need to be post-processed (bit-shifted).\n\n");
              break;
            default:
              DemoLogf("Error: Valid choices for subchannel Raw Data capacity are 8, 16, 32\n");
              DemoHelpPanel();
              OsaConsoleExit();
              exit(1);
            }
          }
      } else if (*(argv[i]+1) == 'M') {   // Li Mode
          if (*(argv[i]+2) == 'E') {   // E1 
            envP->eT1E1Mode    = OTX_T1E1_MODE_E1;
          } else if (*(argv[i]+2) == 'T') {   // T1
            envP->eT1E1Mode    = OTX_T1E1_MODE_T1;
          } else {
            DemoLogf("Error: Invalid Li Mode (E1/T1) command line switch %s\n", argv[i]);
            DemoHelpPanel();
            OsaConsoleExit();
            exit(1);
          }
        } else {
          DemoLogf("Error: Unknown command line switch %s\n", argv[i]);
          DemoHelpPanel();
          OsaConsoleExit();
          exit(1);
        }
      } else {
        DemoLogf("Error: Unknown command line switch %s\n", argv[i]);
        DemoHelpPanel();
        OsaConsoleExit();
        exit(1);
      }
    }
  }
}


/*******************************************************************************

  @func

  Displays a help for the command line.

  @comm

*******************************************************************************/
static void DemoHelpPanel()
{
  DemoLogf("\nCommand line switches:\n");
  DemoLogf(" /L<fileName> : Log file name\n");
  DemoLogf(" /B<boardNo>  : Board number (Default is %d)\n", DEMO_DEFAULT_BOARD_NO);
  DemoLogf(" /T[P|S]      : Board Type used for the demo, where\n");
  DemoLogf("                  P = Use Thor-8-PCI-Plus Board\n");
  DemoLogf("                  S = Use Sleipnir Board\n");
  DemoLogf(" /ME          : E1 Li mode\n");
  DemoLogf(" /MT          : T1 Li mode\n");
  DemoLogf(" /N<capacity> : Raw Data Pipe capacity in kbps (in 64kbps increments)\n");
  DemoLogf(" /S<timeslot> : Starting timeslot (default is TS1)\n");
  DemoLogf(" /Q           : Exit without prompting on errors\n");
}


/*******************************************************************************

  @func

  Displays a message on the screen and prints it to the logfile.

  @comm

*******************************************************************************/
static void DemoLogf(char *formatStr, ...)
{
  va_list argList;

  //EnterCriticalSection(&g_demoLogfCritSec);

  va_start(argList, formatStr);

  OsaConsolePrintF(formatStr, argList);
  if (g_pDemoLogFile) vfprintf(g_pDemoLogFile,   formatStr, argList);

  va_end(argList);

  //LeaveCriticalSection(&g_demoLogfCritSec);
}


/*******************************************************************************
*******************************************************************************/
void DemoAtExit(void)
{
  OTX_INT32 nLi;

  fclose(g_pDemoLogFile);


  for (nLi = 0; nLi < g_demoEnv.coLi; nLi++) {
    if (g_hLogicDevs.mRawDataSender[nLi].pTxFile) {
      fclose(g_hLogicDevs.mRawDataSender[nLi].pTxFile);
      g_hLogicDevs.mRawDataSender[nLi].pTxFile = NULL;
    }
    if (g_hLogicDevs.mRawDataReceiver[nLi].pRxFile) {
      fclose(g_hLogicDevs.mRawDataReceiver[nLi].pRxFile);
      g_hLogicDevs.mRawDataReceiver[nLi].pRxFile = NULL;
    }
  }
}

int main(int argc, char *argv[])
/*******************************************************************************
*******************************************************************************/
{
  OTX_INT32 exitFlag = 0;
  OTX_RESULT result;  // OTX Standard Return Code
  char commandChar;

	OsaConsoleInit();

  printf("T8PBURST, %s, Odin TeleSystems Inc.\n", __DATE__);
  printf("Example program for Burst device.\n");

  DemoParseCmdLineArgs(argc, argv, &g_demoEnv);

  if (g_demoEnv.pszInputStr) {
    OsaConsoleInput(g_demoEnv.pszInputStr);
  }

  if ((g_pDemoLogFile = fopen(g_demoEnv.logFileName, "wt")) == NULL) {
    printf("Couldn't open logfile %s\n", g_demoEnv.logFileName);
    OsaConsoleExit();
    exit(-1);
  }
  atexit(DemoAtExit);

  {
    OTX_INT32 nMaxCapacity;

    if (g_demoEnv.eT1E1Mode == OTX_T1E1_MODE_T1) {
      nMaxCapacity = 1536;
    } else {
      nMaxCapacity = 1984;
    }
    
    if (g_demoEnv.nRawDataCapacity > nMaxCapacity) {
      DemoLogf("Error: Capacity %d exceeded maximum of %d.\n", g_demoEnv.nRawDataCapacity, nMaxCapacity);
      OsaConsoleExit();
      OtxDrvReleaseLib();
      exit(1);
    }
  }

  DemoLogf("Log file is %s\n", g_demoEnv.logFileName);
  DemoLogf("Raw Data Pipe capacity is %d\n", g_demoEnv.nRawDataCapacity);

  // Query Library revision
  {
    char szVersionInfo[255];
    OTX_BOOL bIsStaticLinkedLibrary;
    char szLibFileName[64];
    result = OtxDrvGetAttributeValueSz(OTX_LIB_HANDLE, OTX_ATTR_LIB_REV_STR, szVersionInfo, 255); DemoCheckRc(OTX_INVALID_HANDLE_VALUE, result);
    result = OtxDrvGetAttributeValueBool(OTX_LIB_HANDLE, OTX_ATTR_LIB_TYPE_STATIC, &bIsStaticLinkedLibrary); DemoCheckRc(OTX_INVALID_HANDLE_VALUE, result);
    result = OtxDrvGetAttributeValueSz(OTX_LIB_HANDLE, OTX_ATTR_LIB_FILE_NAME, szLibFileName, 64); DemoCheckRc(OTX_INVALID_HANDLE_VALUE, result);
    
    DemoLogf("Library version %s (filename %s, %s linked)\n", szVersionInfo, szLibFileName, bIsStaticLinkedLibrary ? "static":"dynamic");
  }


  // Connect to the Driver
  DemoLogf("Connecting lib...\n");
  result = OtxDrvConnectLib(); 
  switch (OTX_RESULT_CODE(result)) {
  case OTX_RES_HANDLE:
    DemoLogf("Error: Could not connect to the OTX library.\nPossible cause: OTX Driver not loaded\n");
    OsaConsoleExit();
    OtxDrvReleaseLib();
    exit(1);
    break;
  case OTX_RES_NOT_SUPPORTED:
    DemoLogf("Error: Mismatch between OTX Driver revision and OTX Library revision.\nRevisions must match.\n");
    OsaConsoleExit();
    OtxDrvReleaseLib();
    exit(1);
    break;
  default:
    DemoCheckRc(OTX_INVALID_HANDLE_VALUE, result);
    break;
  }

  // Create Event Queues for Driver --> Application Communications
  DemoLogf("DemoCreateEventQueue\n");
  DemoCreateEventQueue();

  // Connect (get handles) to the physical devices on the Thor-2-PCMCIA board
  DemoLogf("DemoCreatePhysicalDevices\n");
  DemoCreatePhysicalDevices();

  // Configure E1/T1 devices
  DemoLogf("Configuring E1/T1 devices...\n");
  DemoConfigureE1T1Devices();
  DemoLogf("Configured E1/T1 devices\n");

  // Enable physical devices. Their device state should now be OTX_DEVICE_WORKING
  DemoLogf("DemoEnablePhysicalDevices\n");
  DemoEnablePhysicalDevices();

  // Set clock source for the board 
  result = OtxBrdSetClocks(g_hPhysDevs.hThorController, OTX_CLOCK_MODE_LOCAL, OTX_CLOCK_SOURCE_INTERNAL); DemoCheckRc(g_hPhysDevs.hThorController, result);

  // Create and get handles to the Logical devices needed by this application
  DemoLogf("DemoCreateLogicalDevices\n");
  DemoCreateLogicalDevices();

  // Create and get handles to the Pipe devices needed by this application
  DemoLogf("DemoConnectDevices\n");
  DemoConnectDevices();

  // Initialize the logical devices. A logical device must be create and connected before it can be initialized.
  DemoLogf("DemoInitializeLogicalDevices\n");
  DemoInitializeLogicalDevices();

  // Enable logical devices. Their device state should now be OTX_DEVICE_WORKING
  DemoLogf("DemoEnableLogicalDevices\n");
  DemoEnableLogicalDevices();

  // Main Loop. Loop here and wait for Events from the driver or commands from the user ....
	OsaThreadStart(DemoIdleFunction);

  OsaThreadSleep(300); // Only so the following message will show on the screen

  DemoLogf("\nPress 'h' for help\n");

  while (!exitFlag) {
    commandChar = OsaConsoleGetCh();
    exitFlag = DemoProcessUserCommand(commandChar);
  }

  // abort threads
  DemoLogf("\nExiting...\n");
  {
    OTX_INT32 nLi;

    for (nLi = 0; nLi < g_demoEnv.coLi; nLi++) {
      g_hLogicDevs.mRawDataReceiver[nLi].bActive = OTX_FALSE;
      g_hLogicDevs.mRawDataSender[nLi].bActive = OTX_FALSE;
    }

    OsaThreadSleep(1000);   // give threads a chance to exit
  }

  // Clean up
  OsaConsoleExit();
  OtxDrvReleaseLib();

  return 0;
}
